/*-------------------------------------------------------------------------+
| Copyright 2016 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.backend;

import static org.fortiss.af3.exploration.alg.util.DesignSpaceExplorationModelElementFactory.createDefaultDesignSpaceExploration;
import static org.fortiss.af3.exploration.alg.util.ExplorationAlgUtils.extractTargetSuperSetFromAllocConstr;
import static org.fortiss.tooling.common.util.LambdaUtils.getFirst;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.getParentsWithType;
import static org.fortiss.tooling.kernel.utils.LoggingUtils.showInfo;

import java.util.Arrays;
import java.util.Collection;
import java.util.HashSet;
import java.util.Optional;
import java.util.Set;

import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.swt.widgets.Display;
import org.fortiss.af3.component.model.ComponentArchitecture;
import org.fortiss.af3.exploration.alg.dse.Explorer;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.Phenotype;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.solution.StrictTTSchedule;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.SystemModelAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.af3.AF3SystemModelAdapter;
import org.fortiss.af3.exploration.alg.feature.DesignDiversityExploration;
import org.fortiss.af3.exploration.alg.feature.IExplorationFeature;
import org.fortiss.af3.exploration.alg.feature.PartitionArchitectureExploration;
import org.fortiss.af3.exploration.alg.feature.SafetyArchitectureExploration;
import org.fortiss.af3.exploration.alg.feature.TTScheduleExploration;
import org.fortiss.af3.exploration.alg.feature.TaskMappingExploration;
import org.fortiss.af3.exploration.alg.plot.XYPlotter;
import org.fortiss.af3.exploration.dsl_v2.model.expression.SuperSet;
import org.fortiss.af3.exploration.model.ExplorationSpecification;
import org.fortiss.af3.exploration.model.SuperSetMap;
import org.fortiss.af3.exploration.model.solutions.ExplorationSolution;
import org.fortiss.af3.exploration.moea.model.DseSpecification;
import org.fortiss.af3.exploration.service.IDSEBackend;
import org.fortiss.af3.platform.model.ExecutionUnit;
import org.fortiss.af3.platform.model.PlatformArchitecture;
import org.fortiss.af3.task.model.Signal;
import org.fortiss.af3.task.model.Task;
import org.fortiss.af3.task.model.TaskArchitecture;
import org.fortiss.af3.task.model.allocation.ComponentToTaskAllocationEntry;
import org.fortiss.tooling.base.model.element.IModelElement;
import org.fortiss.tooling.kernel.extension.data.ITopLevelElement;
import org.fortiss.tooling.kernel.service.IPersistencyService;
import org.jfree.ui.RefineryUtilities;

/**
 * Contributor class to allow launching this DSE backend from the DSE perspective.
 * 
 * @author diewald
 */
public class Opt4JDseBackend implements IDSEBackend {

	/** Specifies the required input models from the DSE. */
	private final Collection<Class<?>> requiredInputModels = Arrays.asList(
			ComponentArchitecture.class, TaskArchitecture.class, PlatformArchitecture.class);

	// TODO: Remove after migration to AF3 visualization
	/** Last {@link XYPlotter} (in order to close it before creating a new one. */
	protected XYPlotter lastPlotter;

	/** {@inheritDoc} */
	@Override
	public Collection<Class<?>> getRequiredInputSet() {
		return requiredInputModels;
	}

	/** {@inheritDoc} */
	@Override
	public Optional<ExplorationSolution> executeDSE(ExplorationSpecification expSpec,
			Set<Class<? extends IModelElement>> solutionModelTypes,
			Collection<org.fortiss.af3.exploration.model.IExplorationFeature> explorationModules,
			IProgressMonitor monitor, int timeoutMS) throws Exception {
		SystemModelAdapter<?, ?, ?, ?, ?, ?, ?> systemModelAdapter;
		SuperSetMap superSets = expSpec.getSearchSpace();
		// TODO: Find a cleaner way to extract the model context.
		TaskArchitecture taskArchitecture =
				getArchModel(superSets, TaskArchitecture.class, Task.class);
		ITopLevelElement context =
				IPersistencyService.getInstance().getTopLevelElementFor(taskArchitecture);

		// TODO: Move to the common exploration plugin as a workaround for now.
		SuperSet<ExecutionUnit> selectedExecSet =
				extractTargetSuperSetFromAllocConstr(superSets.get(Task.class), expSpec);
		if(selectedExecSet == null) {
			showInfo("No allocation constraint" +
					" on all Tasks was set. Taking all existing ExecutionUnits as targets" +
					" for now. This will not work with hierarchical platforms.");
			selectedExecSet = superSets.get(ExecutionUnit.class);
		}

		// Input to DSE: {@link SystemModelAdapter} and {@link DesignSpaceExploration}.
		// SystemModelAdapter systemModelAdapter;
		systemModelAdapter = new AF3SystemModelAdapter(superSets.get(Task.class),
				superSets.get(ComponentToTaskAllocationEntry.class), selectedExecSet,
				superSets.get(Signal.class));
		DseSpecification dseSpec = createDefaultDesignSpaceExploration(expSpec, context);
		addDefaultExplorationFeatures(dseSpec);

		// Set the requested solutions.
		Set<Class<? extends Phenotype>> solutionTypes = new HashSet<>();
		solutionTypes.add(StrictTTSchedule.class);

		try {
			Explorer explorer = new Explorer();
			// Set<Class<? extends Phenotype>> solPhenotypeTypes = new HashSet<>();
			for(Class<? extends Phenotype> solType : solutionTypes) {
				// TODO: Shall we keep this, and warn about missing Genotyped annotations?
				if(!Phenotype.class.isAssignableFrom(solType)) {
					throw new Exception(
							"The demanded solution is not available in the sleceted DSE backend.");
				}
			}
			ExplorationSolution explorationSolution =
					explorer.explore(dseSpec, systemModelAdapter, solutionTypes, monitor, true);

			Display.getDefault().asyncExec(() -> plotResult(explorationSolution));
			return Optional.of(explorationSolution);
		} catch(Exception ex) {
			// TODO: readd code to present the schedule which failed... Use the LoggingService from
			// the kernel.
			// ExplorationUiUtil.showExceptionErrorAsync(
			// "Design Space Exploration: Error", ex.getMessage(), ex);
			// if(ex instanceof ScheduleRuntimeException) {
			// StrictTTSchedule<?, ?> faultySchedule =
			// ((ScheduleRuntimeException)ex).getSchedule();
			// SchedulePlotter sp = new SchedulePlotter(faultySchedule);
			// sp.createPlot();
			// }
			throw ex;
		}
	}

	/**
	 * Temporary helper method to extract the architecture model from a {@link SuperSet} provided in
	 * the {@link SuperSetMap} to the DSE.
	 */
	private <T extends IModelElement, S extends IModelElement> T getArchModel(SuperSetMap superSets,
			Class<T> archModelType, Class<S> elementType) throws Exception {
		Optional<S> anyElem = getFirst(superSets.get(elementType).getEntries());
		if(!anyElem.isPresent()) {
			throw new Exception("No " + elementType.getSimpleName() +
					" has been passed to the DSE in the corresponding set. Exiting.");
		}
		return getFirst(getParentsWithType(anyElem.get(), archModelType)).orElse(null);
	}

	/**
	 * Adds the default set of {@link IExplorationFeature}s to the given {@link DseSpecification}.
	 */
	private void addDefaultExplorationFeatures(DseSpecification dseSpec) {
		dseSpec.getFeatures().add(SafetyArchitectureExploration.class);
		dseSpec.getFeatures().add(PartitionArchitectureExploration.class);
		dseSpec.getFeatures().add(TaskMappingExploration.class);
		dseSpec.getFeatures().add(DesignDiversityExploration.class);
		dseSpec.getFeatures().add(TTScheduleExploration.class);
	}

	/**
	 * Creates and launches the Dialog that displays the results from the exploration.
	 * Pareto-optimal points are displayed in a 2D-graph, while schedules can be inspected for
	 * each solution.
	 */
	@Override
	public void plotResult(ExplorationSolution solutionSet) {
		if(solutionSet != null) {
			// Keep only one DSE result window
			if(lastPlotter != null) {
				lastPlotter.dispose();
			}

			// draw figure
			// FIXME: Unchecked cast to ExplorationSolutionSet
			XYPlotter plotter = new XYPlotter(solutionSet,
					"Design Space Exploration Result (generations: N/A)");
			plotter.createPlot();
			plotter.pack();
			RefineryUtilities.centerFrameOnScreen(plotter);
			plotter.setVisible(true);

			lastPlotter = plotter;
		}
	}

	/** {@inheritDoc} */
	@Override
	public String getName() {
		return "MOEA-based Exploration";
	}

	/** {@inheritDoc} */
	@Override
	public String toString() {
		return getName();
	}
}
