/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse;

import static org.fortiss.af3.exploration.alg.dse.ImplicitExplorationTargetFactory.createPartitionMinObjective;
import static org.fortiss.af3.exploration.alg.service.ExplorationServiceManager.getService;
import static org.fortiss.tooling.common.util.LambdaUtils.filter;
import static org.fortiss.tooling.common.util.LambdaUtils.filterType;

import java.util.Collection;
import java.util.Collections;
import java.util.Set;

import org.eclipse.core.runtime.IProgressMonitor;
import org.fortiss.af3.exploration.alg.dse.backend.IExplorerBackend;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.Opt4JExplorerBackend;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.DseInjectorProvider;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.create.CreateInjectorProvider;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.create.CreateInterceptModule;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.create.CreateInterceptor;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.decode.DecodeInjectorProvider;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.decode.DecodeInterceptor;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.decode.DecodingInterceptModule;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.evaluate.EvaluatorInjectorProvider;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.Phenotype;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.thread.MultiThreadingModule;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.SystemModelAdapter;
import org.fortiss.af3.exploration.alg.service.IExplorationConstraintTransformationService;
import org.fortiss.af3.exploration.model.ExplorationConstraint;
import org.fortiss.af3.exploration.model.ExplorationSpecification;
import org.fortiss.af3.exploration.model.ExplorationTarget;
import org.fortiss.af3.exploration.model.solutions.ExplorationSolution;
import org.fortiss.af3.exploration.model.solutions.SingleExplorationSolution;
import org.fortiss.af3.exploration.moea.model.DseSpecification;
import org.fortiss.af3.exploration.moea.model.parameters.EAOptimizer;
import org.fortiss.af3.exploration.moea.model.parameters.IOptimizer;
import org.fortiss.tooling.kernel.utils.EcoreUtils;

import com.google.inject.Guice;
import com.google.inject.Injector;

/**
 * Explorer - entry point to architectural exploration
 * 
 * @author huang
 */
public class Explorer {

	/**
	 * Performs the DSE defined a given {@link DseSpecification} specification on a given
	 * system model (represented by a {@link SystemModelAdapter}).
	 * 
	 * @throws Exception
	 */
	public ExplorationSolution explore(DseSpecification dse,
			SystemModelAdapter<?, ?, ?, ?, ?, ?, ?> systemModelAdapter,
			Set<Class<? extends Phenotype>> requestedSolutions, boolean generateImplicitConstraints)
			throws Exception {
		return explore(dse, systemModelAdapter, requestedSolutions, null,
				generateImplicitConstraints);
	}

	/**
	 * Performs the DSE defined a given {@link DseSpecification} specification on a given
	 * system model (represented by a {@link SystemModelAdapter}).
	 */
	public ExplorationSolution explore(DseSpecification dseSpec,
			SystemModelAdapter<?, ?, ?, ?, ?, ?, ?> systemModelAdapter,
			Set<Class<? extends Phenotype>> requestedSolutions, IProgressMonitor progressMonitor,
			boolean generateImplicitConstraints) throws Exception {
		validateDesignSpaceExploration(dseSpec);
		validateSystemModelAdapter(systemModelAdapter);

		// TODO: Remove this hack to add the partition minimization objective to the original target
		// spec
		ExplorationSpecification origExpSpec = dseSpec.getTargetSpecification();
		origExpSpec.getTargets().add(createPartitionMinObjective());

		ExplorationSpecification expSpec = EcoreUtils.copy(origExpSpec);
		dseSpec.setTargetSpecification(expSpec);
		if(generateImplicitConstraints) {
			expSpec.getTargets().addAll(ImplicitExplorationTargetFactory
					.createImplicitExplorationTargets(dseSpec, systemModelAdapter));
		}

		// Extract external models to add them to the available input models.
		// TODO: readd support for "external" EMF models
		// for(ExplorationTarget<?> expTarget : expSpec.getTargets()) {
		// if(expTarget instanceof ExternalModelTarget) {
		// Collection<EObject> extModels =
		// ((ExternalModelTarget)expTarget).getExternalModels();
		// for(EObject extModel : extModels) {
		// @SuppressWarnings("unchecked") Class<? extends EObject> modelType =
		// (Class<? extends EObject>)extModel.eClass().getInstanceClass();
		// }
		// }
		// }

		// Execute the Constraint Transformation.
		IExplorationConstraintTransformationService constrService =
				getService(IExplorationConstraintTransformationService.class);
		Collection<ExplorationConstraint<?>> genConstrs =
				filter(filterType(expSpec.getTargets(), ExplorationConstraint.class),
						t -> (t.getExpression() == null));
		// TODO: Reenable the constraint transformation service.
		genConstrs = Collections.emptySet();
		Collection<ExplorationConstraint<?>> transformedConstrs =
				constrService.getTransformedConstraints(systemModelAdapter, genConstrs);
		expSpec.getTargets().addAll(transformedConstrs);

		// Invoke the backend in order to perform the DSE
		IExplorerBackend<? extends IOptimizer> explorerBackend = createExplorerBackend(dseSpec,
				expSpec, systemModelAdapter, requestedSolutions, progressMonitor);
		ExplorationSolution explorationResult = explorerBackend.explore();
		// We use the original ExplorationSpecification for the returned result.
		explorationResult.setExplorationSpec(origExpSpec);

		// Remove the implicit ExplorationTargets from the metrics of each single solution.
		removeImplicitTargets(explorationResult);

		return explorationResult;
	}

	/** Helper method to issue an exception for invalid DSE configurations */
	private void createDesignSpaceExplorationConfigurationException(final String msg)
			throws Exception {
		throw new Exception("DSE configuration: " + msg);
	}

	// TODO: Port validation of DSE configuration to model validation framework that creates markers
	// in the problem view
	/** Validation of DSE configuration */
	private void validateDesignSpaceExploration(DseSpecification dse) throws Exception {
		if(dse == null) {
			createDesignSpaceExplorationConfigurationException(
					"No DSE configuration has been specified.");
			// Never here
			return;
		}

		if(dse.getParameters() == null) {
			createDesignSpaceExplorationConfigurationException(
					"No optimizer independent parameters have been provided.");
		}

		if(dse.getParameters().getExecutionModel() == null) {
			createDesignSpaceExplorationConfigurationException(
					"No execution model has been specified.");
		}

		if(dse.getOptimizer() == null) {
			createDesignSpaceExplorationConfigurationException(
					"No optimizer backend has been specified.");
		}

		// TODO: create a validation whether the constraints or objectives are not given
		if(dse.getTargetSpecification() == null ||
				dse.getTargetSpecification().getTargets() == null ||
				dse.getTargetSpecification().getTargets().isEmpty()) {
			createDesignSpaceExplorationConfigurationException(
					"No Optimization goals have been provided.");
		}
	}

	// TODO: Use a subclassed Exception here.
	/** Helper method to issue an exception for invalid input system models */
	private void createSystemModelAdapterConfigurationException(final String msg) throws Exception {
		throw new Exception("System model: " + msg);
	}

	// TODO: Port validation of system model to model validation framework that creates markers
	// in the problem view
	/** Validation of system model */
	private void validateSystemModelAdapter(
			SystemModelAdapter<?, ?, ?, ?, ?, ?, ?> systemModelAdapter) throws Exception {

		if(systemModelAdapter == null) {
			createSystemModelAdapterConfigurationException(
					"No DSE configuration has been specified.");
			return;
		}

		if(systemModelAdapter.getDeployableComponents().isEmpty()) {
			createSystemModelAdapterConfigurationException(
					"Component architecture does not contain any deployable components.\n\nMake sure the \"Trigger\" annotation have been defined for all components that should be deployed.");
		}

		if(systemModelAdapter.getDeploymentTargets().isEmpty()) {
			createSystemModelAdapterConfigurationException(
					"Platform architecture does not contain any deployment targets.\n\nMake sure the \"Deployment Target\" annotation has been set for all execution units that should be deployed.\n\nMaybe you selected the wrong platform architecture?");
		}

		if(systemModelAdapter.getTaskGraphs().isEmpty()) {
			createSystemModelAdapterConfigurationException(
					"System model does not contain any valid task graphs. All tasks in a task graph must share the same period.");
		}

		if(systemModelAdapter.getAcyclicTaskGraphs().isEmpty()) {
			createSystemModelAdapterConfigurationException(
					"System model does not contain any valid task graphs. All tasks in a task graph must share the same period.");
		}

		// FIXME: readd? ==>> VL Generation.
		// if(systemModelAdapter.getHyperPeriod() <= 0) {
		// createSystemModelAdapterConfigurationException("System model does not have a valid hyper
		// period.");
		// }
	}

	/**
	 * Removes the implicit {@link ExplorationTarget}s that were added internally. They shall not be
	 * visible to the user.
	 */
	private void removeImplicitTargets(ExplorationSolution explorationSolution) {
		ExplorationSpecification origExpSpec = explorationSolution.getExplorationSpec();
		for(SingleExplorationSolution singleSol : explorationSolution.getSolutions()) {
			for(ExplorationTarget<?> explorationTarget : singleSol.keySet()) {
				if(!origExpSpec.getTargets().contains(explorationTarget)) {
					singleSol.removeKey(explorationTarget);
				}
			}
		}
	}

	/** Explorer backend */
	private IExplorerBackend<? extends IOptimizer> createExplorerBackend(DseSpecification dse,
			ExplorationSpecification expSpec,
			SystemModelAdapter<?, ?, ?, ?, ?, ?, ?> systemModelAdapter,
			Set<Class<? extends Phenotype>> requestedSolutions, IProgressMonitor progressMonitor)
			throws Exception {
		InternalSpecificationProvider specProvider =
				new InternalSpecificationProvider(expSpec, requestedSolutions);

		DseInjectorProvider provider = new DseInjectorProvider();
		Injector dseInjector = Guice.createInjector(systemModelAdapter, specProvider,
				new CreateInterceptModule(new CreateInterceptor()), new CreateInjectorProvider(),
				new DecodingInterceptModule(new DecodeInterceptor()), new DecodeInjectorProvider(),
				new EvaluatorInjectorProvider(), new MultiThreadingModule(), provider);
		provider.setInjector(dseInjector);

		if(dse.getOptimizer() instanceof EAOptimizer) {
			return new Opt4JExplorerBackend(dse, expSpec, dseInjector, requestedSolutions,
					progressMonitor);
		}

		createDesignSpaceExplorationConfigurationException("Unsupported exploration backend.");

		// Never here
		return null;
	}
}
