/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.evaluator.objective;

import java.text.DecimalFormat;
import java.text.NumberFormat;
import java.util.HashMap;
import java.util.Map;

import org.apfloat.Apfloat;
import org.apfloat.FixedPrecisionApfloatHelper;
import org.eclipse.emf.ecore.EObject;
import org.fortiss.af3.component.model.Component;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.instantiatedmapping.InstantiatedTaskMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.instantiatedmapping.InstantiatedTaskMappingEntry;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.solution.StrictTTSchedule;
import org.fortiss.af3.exploration.alg.dse.module.EvaluatorModule;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IExecutionUnitAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ITaskAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.mappers.TaskMapper;
import org.fortiss.af3.exploration.moea.model.objectives.FailureMinObjective;

import com.google.inject.Inject;

/**
 * The reliability analysis computes the system-level reliability (in terms of failure probability)
 * based on a given component-level reliability.
 * 
 * @author huang
 * @author diewald
 */
public abstract class ReliabilityEvaluatorBase<S extends InstantiatedTaskMappingEntry, T extends InstantiatedTaskMappingEncoding<S>>
		extends EvaluatorModule<Double, FailureMinObjective, StrictTTSchedule<S, T>> {

	/** Precision of floating point values */
	protected final static int FP_PRECISION = 50;

	/** Component reliability */
	private Map<ITaskAdapter<?>, Map<IExecutionUnitAdapter<?>, Apfloat>> componentReliability =
			new HashMap<ITaskAdapter<?>, Map<IExecutionUnitAdapter<?>, Apfloat>>();

	/** Helper required to perform {@link org.apfloat.Apfloat}-based calculations. */
	private FixedPrecisionApfloatHelper apfloatHelper =
			new FixedPrecisionApfloatHelper(FP_PRECISION);

	private TaskMapper<EObject> taskMapper;

	/**
	 * References the "target" {@link Component} in the original input model. This is the
	 * {@link Component} for which the failures shall be minimized.
	 */
	protected ITaskAdapter<?> targetTask;

	/** Constructor. */
	// TODO: This constructor shall go to the decoding phase since it actually generates a new
	// mapping and task graph.
	@Inject
	public ReliabilityEvaluatorBase(FailureMinObjective target) throws Exception {
		super(target);

		// this.taskMapper = (TaskMapper<EObject>)taskMapper;
		//
		// targetTask = this.taskMapper.getAdapter(explorationTarget.getTargetComponent());
		// if(targetTask == null) {
		// throw new Exception("The Component for which to minimize the failure rate is not" +
		// " correctly specified. It is either not defined, or it is not part of the" +
		// " set of deployable Components. The error is caused by " +
		// explorationTarget.getName() + "(id: " + explorationTarget.getId() + ").");
		// }
		//
		// // TODO: Check for different implementation of this single initialization for different
		// // reliability goals. In particular, it must be checked if the current implementation
		// with a
		// // dedicated ReliabilityEvaluatorBase instance per ReliabilityObjective has the same
		// // functionality and performance as the original implementation.
		// //
		// // If needed, a cache could be implemented. See also similar comment in
		// // FaultDetectionVotingReliabilityEvaluator.evaluateInternal().
		//
		// // do initialization
		// System.out.println("initialize reliability analyzer.");
		// componentReliability.clear();
		//
		// // extract the detector annotations
		// for(ITaskAdapter<?> task : tasks) {
		// TaskGraph tg = tgEnc.getTaskGraphOf(task);
		// // TODO: re-add the consideration of fault detector annotations here after the
		// // annotations (or specifications) are defined
		// // This block adds a predefined fault detector that has has a specified delay
		// // for(IModelElementSpecification spec : comp.getSpecifications()) {
		// // if(spec instanceof FaultDetectorAnnotation) {
		// // // extract the fault detector annotation
		// // FaultDetectorAnnotation annotation = (FaultDetectorAnnotation)spec;
		// // FaultDetector fd =
		// // new FaultDetector(annotation.getOverhead(), annotation.getCoverage(),
		// // annotation.getPath());
		// // tg.addFaultDetector(deployableComponent, fd);
		// // }
		// // }
		// // add the default fault detector with 0 overhead (native execution)
		// tg.addFaultDetector(task, new FaultDetector(0, 0));
		// }
		//
		// // we use the classic poisson model and compute the constant failure rate of each
		// software
		// // task
		// Set<IExecutionUnitAdapter<?>> targetsToBeRemoved = new
		// HashSet<IExecutionUnitAdapter<?>>();
		//
		// for(ITaskAdapter<?> deployableComponent : replicableTasks) {
		//
		// // compute the component failure rates
		// Map<IExecutionUnitAdapter<?>, Apfloat> taskmap =
		// componentReliability.get(deployableComponent);
		// if(taskmap == null) {
		// taskmap = new HashMap<IExecutionUnitAdapter<?>, Apfloat>();
		// componentReliability.put(deployableComponent, taskmap);
		// }
		// // for each core that may execute the task, compute the reliability
		// for(IExecutionUnitAdapter<?> deploymentTarget : deploymentTargets) {
		// double coreFailureRate = deploymentTarget.getFailureRate();
		//
		// if(coreFailureRate < 0) {
		// // this core is not supported, do not map to this core
		// targetsToBeRemoved.add(deploymentTarget);
		// continue;
		// }
		//
		// storeComponentFailureRate(taskmap, deployableComponent, deploymentTarget,
		// coreFailureRate);
		// // } else {
		// // // the resource is not a component but an allocation of a resource
		// // // we look into its parents to look for the reliability annotation
		// // TODO: re-add the consideration of fault detector annotations here after the
		// // annotations (or specifications) are defined.
		// // The iteration over Eobject's containers can likely be avoided.
		// // EObject obj = (EObject)deploymentTarget;
		// // ReliabilityAnnotation annotation = null;
		// // while(obj != null) {
		// // if(obj instanceof AcrossComponent) {
		// // for(IModelElementSpecification spec : ((AcrossComponent)obj)
		// // .getSpecifications()) {
		// // if(spec instanceof ReliabilityAnnotation) {
		// // annotation = (ReliabilityAnnotation)spec;
		// // break;
		// // }
		// // }
		// // // break if we already found the annotation
		// // if(annotation != null)
		// // break;
		// // }
		// // obj = obj.eContainer();
		// // }
		// // if(annotation != null) {
		// // double coreFailureRate = toFailureRate(annotation);
		// // storeComponentFailureRate(taskmap, task, deploymentTarget, coreFailureRate);
		// // } else {
		// // // not supported
		// // targetsToBeRemoved.add(deploymentTarget);
		// // }
		// // storeComponentFailureRate(taskmap, deployableComponent, deploymentTarget,
		// // coreFailureRate);
		// // }
		// }
		// }
	}

	// /**
	// * Determine the {@link FailureProbabilityType} based on the concrete
	// * {@link ReliabilityMaxObjective}, and partially also based on the selected {@link
	// FailureMode}
	// * .
	// */
	// private FailureProbabilityType getFailureProbabilityType(ReliabilityMaxObjective goal) {
	// if(goal instanceof ReliabilityObjectiveDUFfit) {
	// return FailureProbabilityType.ALL;
	// } else if(goal instanceof ReliabilityObjectiveSDCfit) {
	// return FailureProbabilityType.SDC;
	// }
	//
	// assert (goal instanceof ReliabilityObjectiveFailureInTime);
	// if(((ReliabilityObjectiveFailureInTime)goal).getMode() == FailureMode.FAIL_OPERATIONAL) {
	// return FailureProbabilityType.ALL;
	// }
	// assert (((ReliabilityObjectiveFailureInTime)goal).getMode() == FailureMode.FAIL_SAFE);
	// return FailureProbabilityType.SDC;
	// }

	// private double toLogScale(double value) {
	// if(value <= 0) {
	// return -100;
	// }
	//
	// return java.lang.Math.log10(value);
	// }

	/** Stores the failure rate of a component to a task map. */
	private void storeComponentFailureRate(Map<IExecutionUnitAdapter<?>, Apfloat> taskmap,
			ITaskAdapter<?> deployableComponent, IExecutionUnitAdapter<?> deploymentTarget,
			double coreFailureRate) {
		// get the execution time from annotation
		// TODO: heterogeneous cores
		// TODO: Scaling of time unit
		double executionTime = deployableComponent.getWcet(deploymentTarget);
		// deployableComponent.getComponentWCET(deploymentTarget);

		if(executionTime <= 0) {
			System.out.println("WARNING: no delay specification found for component " +
					deployableComponent.getName() + "! set to default 1 primary time unit!");
			executionTime = 1;
		}

		// compute the reliability according to the exponential model
		Apfloat reliability = exponentialReliabilityModel(coreFailureRate, executionTime);
		taskmap.put(deploymentTarget, reliability);
		Apfloat one = new Apfloat(1, FP_PRECISION);
		NumberFormat formatter = new DecimalFormat("0.######E0");
		System.out.println(
				"Task " + deployableComponent.getName() + " on " + deploymentTarget.getName() +
						" : " + formatter.format(one.subtract(reliability).doubleValue()));
	}

	/**
	 * Computes the failure probability of a task w.r.t. the exponential reliability model.
	 * 
	 * @param coreFailureRate
	 *            Execution units failure rate
	 * @param executionTime
	 *            Task's execution time.
	 * @return {@link Apfloat} value of the tasks' failure probability.
	 */
	private Apfloat exponentialReliabilityModel(double coreFailureRate, double executionTime) {
		Apfloat zero = new Apfloat(0, FP_PRECISION);
		Apfloat lambda = new Apfloat(coreFailureRate, FP_PRECISION);
		Apfloat t = new Apfloat(executionTime, FP_PRECISION);

		Apfloat retval = apfloatHelper.exp(zero.subtract(lambda.multiply(t)));

		return retval;
	}

	/** Returns the failure rate for a component. */
	protected Apfloat getComponentFailureRate(ITaskAdapter<?> deployableComponent,
			IExecutionUnitAdapter<?> deploymentTarget) {
		// TODO: This method has to be reconsidered w.r.t. task instantiation and replication.
		if(componentReliability.get(deployableComponent) != null &&
				componentReliability.get(deployableComponent).get(deploymentTarget) != null) {
			Apfloat one = new Apfloat(1, FP_PRECISION);
			return one
					.subtract(componentReliability.get(deployableComponent).get(deploymentTarget));
		}
		return null;
	}
}
