/*-------------------------------------------------------------------------+
| Copyright 2016 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.modeltransformation.base;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.util.EcoreUtil;
import org.fortiss.af3.exploration.alg.dse.modeltransformation.ITransformationModule;
import org.fortiss.af3.exploration.alg.exception.TransformationModuleException;
import org.fortiss.af3.exploration.alg.guava.HashTypedBiMap;
import org.fortiss.af3.exploration.alg.guava.TypedBiMap;
import org.fortiss.tooling.kernel.introspection.IIntrospectionDetailsItem;
import org.fortiss.tooling.kernel.introspection.IIntrospectionItem;

import com.google.common.collect.HashMultimap;
import com.google.common.collect.Multimap;

/**
 * Base class for {@link ITransformationModule}s that maintains a collection of registered
 * submodules.
 * 
 * @author diewald
 */
public abstract class TransformationModuleBase<T extends EObject>
		implements ITransformationModule<T> {

	/**
	 * Holds associations between elements comprising an exploration encoding and their
	 * corresponding model elements.
	 */
	protected TypedBiMap<Object, EObject> encEntryModelElemAssoc = HashTypedBiMap.create();

	/**
	 * Associative map which relates elements of a generated model to those of the original input
	 * model.
	 */
	protected Map<EObject, EObject> genToOrigAssocMap = new HashMap<>();

	/**
	 * Associative map which relates the elements of the original input model to all elements which
	 * from instances of the particular elements.
	 */
	protected Multimap<EObject, EObject> origToGenAssocMap = HashMultimap.create();

	/**
	 * Associative map which relates the input elements of the current stage to the output elements
	 * of the current stage. Here, the current stage does not differentiate between the processing
	 * and the finish stage.
	 */
	protected Multimap<EObject, EObject> genToGenAssocMap = HashMultimap.create();

	/** Maintains references to the submodules of {@code this} {@link ITransformationModule}. */
	protected List<ITransformationModule<?>> subModules = new LinkedList<>();

	/**
	 * Set of sub {@link ITransformationModule}s that shall not be executed in one particular run.
	 */
	private Collection<Class<? extends ITransformationModule<?>>> blacklistedModules =
			new HashSet<>();

	/** {@inheritDoc} */
	@Override
	public List<ITransformationModule<? extends EObject>> getSubModules() {
		return subModules;
	}

	/** {@inheritDoc} */
	@Override
	public void registerSubModule(ITransformationModule<?> subModule) {
		subModules.add(subModule);
	}

	/** {@inheritDoc} */
	@SuppressWarnings("unchecked")
	@Override
	public ITransformationModule<T> createModuleInstance() throws TransformationModuleException {
		Constructor<ITransformationModule<T>> transCtor = null;
		ITransformationModule<T> moduleInstance = null;
		try {
			transCtor = (Constructor<ITransformationModule<T>>)getClass().getDeclaredConstructor();
			moduleInstance = transCtor.newInstance();
		} catch(NoSuchMethodException | SecurityException | IllegalArgumentException |
				InvocationTargetException | InstantiationException | IllegalAccessException e) {
			throw new TransformationModuleException(this,
					"Could not create an instance of this module.", e);
		}

		List<ITransformationModule<?>> subModuleInstances = new ArrayList<>();
		for(ITransformationModule<?> subModule : getSubModules()) {
			ITransformationModule<?> subModuleInst = subModule.createModuleInstance();
			subModuleInstances.add(subModuleInst);
		}
		moduleInstance.getSubModules().clear();
		moduleInstance.getSubModules().addAll(subModuleInstances);

		return moduleInstance;
	}

	/** {@inheritDoc} */
	@Override
	public void applyAssociationUpdates(ITransformationModule<?> subModule) {
		this.encEntryModelElemAssoc = subModule.getEncEntryModelElemAssoc();
		this.origToGenAssocMap = subModule.getOrigToGenAssocMap();
		this.genToOrigAssocMap = subModule.getGenToOrigAssocMap();
		this.genToGenAssocMap = subModule.getGenToGenAssocMap();
	}

	/** {@inheritDoc} */
	@Override
	public EObject getModelElementByEncEntry(Object encEntry, Class<EObject> type) {
		return encEntryModelElemAssoc.get(encEntry, type);
	}

	/** {@inheritDoc} */
	@Override
	public EObject getOriginalByGenElement(EObject genElem) {
		return genToOrigAssocMap.get(genElem);
	}

	/** {@inheritDoc} */
	@Override
	// TODO: rename to removeGenModelElement.
	public void removeGenModelElement(EObject element) {
		removeModelElementAssociations(element);
		EcoreUtil.delete(element);
	}

	/** Removes the given {@code element} from the association maps. */
	private void removeModelElementAssociations(EObject element) {
		for(EObject curChildElem : element.eContents()) {
			removeModelElementAssociations(curChildElem);
		}
		EObject origElement = genToOrigAssocMap.get(element);
		origToGenAssocMap.remove(origElement, element);
		if(origToGenAssocMap.get(origElement) != null &&
				origToGenAssocMap.get(origElement).isEmpty()) {
			// To avoid list with empty associations.
			origToGenAssocMap.removeAll(origElement);
		}
		genToOrigAssocMap.remove(element);
	}

	/** {@inheritDoc} */
	@Override
	public void addBlacklistedModule(Class<? extends ITransformationModule<?>> subModule) {
		blacklistedModules.add(subModule);
	}

	/** {@inheritDoc} */
	@Override
	public void resetBlacklistedModules() {
		blacklistedModules.clear();
	}

	/** {@inheritDoc} */
	@Override
	public boolean showInIntrospectionNavigation() {
		return true;
	}

	/** {@inheritDoc} */
	@Override
	public String getIntrospectionLabel() {
		return getClass().getSimpleName();
	}

	/** {@inheritDoc} */
	@Override
	public Collection<IIntrospectionItem> getIntrospectionItems() {
		return null;
	}

	/** {@inheritDoc} */
	@Override
	public IIntrospectionDetailsItem getDetailsItem() {
		return null;
	}

	/** {@inheritDoc} */
	@Override
	public TypedBiMap<Object, EObject> getEncEntryModelElemAssoc() {
		return encEntryModelElemAssoc;
	}

	/** {@inheritDoc} */
	@Override
	public Map<EObject, EObject> getGenToOrigAssocMap() {
		return genToOrigAssocMap;
	}

	/** {@inheritDoc} */
	@Override
	public Multimap<EObject, EObject> getOrigToGenAssocMap() {
		return origToGenAssocMap;
	}

	/** {@inheritDoc} */
	@Override
	public Multimap<EObject, EObject> getGenToGenAssocMap() {
		return genToGenAssocMap;
	}
}
