/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.sysmodel.arch;

import java.util.Collection;

import org.eclipse.emf.common.util.Enumerator;
import org.fortiss.af3.component.model.Component;

/**
 * Adapter between the deployable components (=tasks) of the system meta-model used in the modeling
 * tool, and the DSE.
 * 
 * @author barner
 */
public interface ITaskAdapter<T> extends IRequestAdapter<T>, Cloneable {
	/** TEMP HACK: Returns the AF3 component realized in this task. */
	// FIXME: Remove after updating the transformation framework for SuperSets.
	Component getAssignedComponent();

	/**
	 * Returns the WCET if this {@link ITaskAdapter} would be executed on the given
	 * {@link IExecutionUnitAdapter}.
	 */
	Double getWcet(IExecutionUnitAdapter<?> executionUnit);

	/**
	 * Returns the consumed energy if this {@link ITaskAdapter} would be executed on
	 * the given {@link IExecutionUnitAdapter}.
	 */
	Double getEnergyConsumption(IExecutionUnitAdapter<?> executionUnit);

	/** Returns the minimum required safety integrity level of the deployable component. */
	Enumerator getRequiredSafetyIntegrityLevel();

	/** Returns the period of the deployable component. */
	Double getPeriod();

	/** Sets the period of the deployable component. */
	void setPeriod(double period);

	/**
	 * Predicates whether this {@link ITaskAdapter} must be instantiated by another
	 * {@link ITaskAdapter}. This is determined by evaluating whether
	 * {@link ITaskAdapter}s are given that may be used to instantiate this
	 * {@link ITaskAdapter}.
	 */
	boolean isInstantiationRequired();

	/**
	 * Returns those {@link ITaskAdapter}s that may be used to instantiate this
	 * {@link ITaskAdapter}.
	 */
	Collection<ITaskAdapter<T>> getReplacementComponents();

	/**
	 * Predicates whether this {@link ITaskAdapter} instantiates an abstract
	 * {@link ITaskAdapter}.
	 */
	boolean isInstantiation();

	/**
	 * Returns the {@link ITaskAdapter} instantiated by this
	 * {@link ITaskAdapter} (if applicable).
	 */
	ITaskAdapter<T> getReplacedComponent();

	/**
	 * Sets the replacedComponent reference of {@code this} {@link ITaskAdapter} to
	 * allow tracking of the "abstract" {@link ITaskAdapter}s from the instantiated
	 * {@link ITaskAdapter}s.
	 */
	// TODO: If possible, find a way to eliminate this suppresswarning.
	void setReplacedComponent(@SuppressWarnings("rawtypes") ITaskAdapter abstractComponent);

	/**
	 * Whenever an {@link ITaskAdapter} is copied/cloned its references wrt. the task
	 * instantiation need to be updated. Hence, after cloning, this method should be called for each
	 * copied/cloned {@link ITaskAdapter}.
	 */
	void updateInstatiationData();

	/**
	 * Returns the minimum number of replications allowed for this
	 * {@link ITaskAdapter}.
	 */
	int getMinReplication();

	/**
	 * Returns the maximum number of replications allowed for this
	 * {@link ITaskAdapter}.
	 */
	int getMaxReplication();

	/**
	 * Clones the {@link ITaskAdapter} on which this method is called. If
	 * {@link Component}s are instantiated during the exploration, this is needed. The initially
	 * constructed {@link ITaskAdapter}s must be replicated for each instantiated
	 * encoding.
	 */
	ITaskAdapter<T> clone();
}
